/*-
 * Copyright (C) 2006 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package org.catacombae.jab;

import org.catacombae.rarx.*;
import org.catacombae.jab.gui.*;
import net.iharder.dnd.FileDrop;
import java.util.Vector;
import java.util.zip.CRC32;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;

public class MainWindow extends JFrame {
    private static String WINDOW_TITLE = "Java Archive Browser";
    private MainPanel mainPanel;
    private RARFileEntry[] entries = new RARFileEntry[0];

    public MainWindow() {
	super(WINDOW_TITLE);
	mainPanel = new MainPanel();
	add(mainPanel, BorderLayout.CENTER);
	
	mainPanel.extractButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent evt) {
		    actionExtract();
		}
	    });
	
	mainPanel.fileEntryTable.setModel(createNewTableModel(mainPanel.fileEntryTable.getModel(), entries));
	//mainPanel.fileEntryTable.setShowGrid(false);
	//mainPanel.fileEntryTableScroller.setBackground(mainPanel.fileEntryTable.getBackground());
	//table.getColumnModel().getColumn(0).setPreferredWidth(200);

	
	mainPanel.fileEntryTable.addComponentListener(new ComponentAdapter() {
		private Dimension oldScrollerSize = null;//new Dimension(0, 0);
		public void componentResized(ComponentEvent ce) {
		    Dimension newScrollerSize = mainPanel.fileEntryTableScroller.getSize();
		    System.out.println("table resized");
		    System.out.println("  table size: " + mainPanel.fileEntryTable.getSize());
		    System.out.println("  scroller size: " + newScrollerSize);
		    if(true || oldScrollerSize == null) {
			Dimension firstScrollerSize = mainPanel.fileEntryTable.getPreferredSize();
			if(firstScrollerSize.width == 0 && firstScrollerSize.height == 0)
			    return;
			else {
			    Insets scrollerInsets = mainPanel.fileEntryTableScroller.getInsets();
			    System.out.println("scrollerInsets: " + scrollerInsets);
			    firstScrollerSize.width += (scrollerInsets.left + scrollerInsets.right);
			    JScrollBar verticalScrollBar = mainPanel.fileEntryTableScroller.getVerticalScrollBar();
			    if(verticalScrollBar != null && verticalScrollBar.isVisible())
				firstScrollerSize.width += verticalScrollBar.getSize().width;
			    oldScrollerSize = firstScrollerSize;
			}
		    }
		    int widthdiff = newScrollerSize.width - oldScrollerSize.width;
		    if(widthdiff != 0) {
			TableColumn firstColumn = mainPanel.fileEntryTable.getColumnModel().getColumn(0);
			firstColumn.setPreferredWidth(firstColumn.getPreferredWidth()+widthdiff);
			mainPanel.fileEntryTable.invalidate();
			mainPanel.fileEntryTable.validate();
			//mainPanel.fileEntryTable.setSize(mainPanel.fileEntryTableScroller.getSize());
			oldScrollerSize = newScrollerSize;
		    }
		}
	    });
	mainPanel.fileEntryTableScroller.addComponentListener(new ComponentAdapter() {
		public void componentResized(ComponentEvent ce) {
		    System.out.println("scroller resized");
		    System.out.println("  table size: " + mainPanel.fileEntryTable.getSize());
		    System.out.println("  scroller size: " + mainPanel.fileEntryTableScroller.getSize());
		}
	    });
	
	new FileDrop(this, new FileDrop.Listener() {
		public void filesDropped(java.io.File[] files) {   
		    // handle file drop
		    if(files.length > 0)
			loadFile(files[0]);
// 		    for(File f : files)
// 			loadFile(f);
		}   // end filesDropped
	    }); // end FileDrop.Listener

	setDefaultCloseOperation(EXIT_ON_CLOSE);
	pack();
	setLocationRelativeTo(null);
    }

    public void actionExtract() {
	final int[] selection = mainPanel.fileEntryTable.getSelectedRows();
	if(selection.length == 0) {
	    JOptionPane.showMessageDialog(this, "Nothing selected.", "Error.", JOptionPane.ERROR_MESSAGE);
	    return;
	}
	JFileChooser jfc = new JFileChooser();
	jfc.setMultiSelectionEnabled(false);
	jfc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
	if(jfc.showSaveDialog(MainWindow.this) == JFileChooser.APPROVE_OPTION) {
	    final File outDir = jfc.getSelectedFile();
	    if(!outDir.isDirectory()) {
		JOptionPane.showMessageDialog(this, "Please select a directory where the unpacked files will be saved.", "Error.", JOptionPane.ERROR_MESSAGE);
		return;
	    }
	    //byte[] buffer = new byte[4096];
	    //CRC32 crc = new CRC32();
	    
	    final ProgressDialog progressDialog = new ProgressDialog(this, true);
	    progressDialog.setMessage("Extracting");
	    progressDialog.setProgress(0.0);
	    final boolean[] breakSignal = { false };
	    progressDialog.addCancelButtonListener(new ActionListener() {
		    public void actionPerformed(ActionEvent ae) {
			breakSignal[0] = true;
			progressDialog.setVisible(false);
		    }
		});
	    new Thread(new Runnable() {
		    public void run() {
			for(int i : selection) {
			    progressDialog.setMessage("Extracting " + entries[i].getHeader(0).getFilenameAsString() + "...");
			    System.err.println("Calling extractToFile...");
			    extractToFile(entries[i], outDir, progressDialog, breakSignal);
			    System.err.println("Call to extractToFile completed.");
			    if(breakSignal[0])
				break;
			}
			progressDialog.setVisible(false);
			
		    }
		}).start();
	    System.err.println("Setting progressDialog visible.");
	    progressDialog.setVisible(true);
	    System.err.println("Done.");
	}
    }

    private void extractToFile(RARFileEntry entry, File outDir, ProgressDialog progressDialog, boolean[] breakSignal) {
	NewFileHeader curHead = entry.getHeader(0);
	if(curHead.getMethod() == RARHeader.METHOD_UNCOMPRESSED) {
	    extractUncompressed(entry, outDir, progressDialog, breakSignal);
	}
	else if(true) {
	    //Let's try decompressing. No idea how this will work.
	    extractCompressed(entry, outDir, progressDialog, breakSignal);
	}
	else
	    JOptionPane.showMessageDialog(MainWindow.this, "Compression method not supported for file \"" +
					  entry.getFile(0) + " (method 0x" +
					  Util.toHexStringBE(curHead.getMethod()) + ").", 
					  "Error.", JOptionPane.ERROR_MESSAGE);
    }
    
    private void extractUncompressed(RARFileEntry entry, File outDir, ProgressDialog progressDialog, boolean[] breakSignal) {
	NewFileHeader curHead = entry.getHeader(0);
	File outFile = new File(outDir, curHead.getFilenameAsString());
	try {
	    mainPanel.statusArea.append("Unstoring to \"" + outFile + "\"...");
	    boolean dirCreated = true;
	    if(!outFile.getParentFile().exists())
		dirCreated = outFile.getParentFile().mkdirs();
	    else if(!outFile.getParentFile().isDirectory())
		dirCreated = false;
	    
	    if(dirCreated) {
		RARFileEntryStream in = new RARFileEntryStream(entry);
		FileOutputStream out = new FileOutputStream(outFile);
		long bytesToRead = entry.getPackedSize();
		long totalBytesRead = 0;
		CRC32 crc = new CRC32();
		byte[] buffer = new byte[4096];
		int bytesRead = in.read(buffer);

		while(bytesRead != -1 && !breakSignal[0]) {
		    totalBytesRead += bytesRead;
		    progressDialog.setProgress(((double)totalBytesRead)/bytesToRead);
		    crc.update(buffer, 0, bytesRead);
		    out.write(buffer, 0, bytesRead);
		    bytesRead = in.read(buffer);
		}
		out.close();
		in.close();
		if(breakSignal[0]) {
		    JOptionPane.showMessageDialog(this, "The file has been partially extracted. Please remove " +
						  "it manually.", "Error.", JOptionPane.ERROR_MESSAGE);
		}
		else {
		    if((int)crc.getValue() == (entry.getFinalFileCRC()/* & 0xFFFFFFFFL*/))
			mainPanel.statusArea.append("done!\n");
		    else {
			JOptionPane.showMessageDialog(MainWindow.this, "CRC-32 check failed while extracting to \"" + 
						      outFile + "\"!\n" +
						      "The file has been extracted but is probably corrupted.\n" + 
						      "Calculated CRC-32: 0x" +
						      Util.toHexStringBE((int)crc.getValue()) + "\n" +
						      "File CRC-32: 0x" + Util.toHexStringBE(curHead.getFileCRC()), 
						      "Error.", JOptionPane.ERROR_MESSAGE);
			mainPanel.statusArea.append("failed!\n");
		    }
		}
	    }
	    else
		JOptionPane.showMessageDialog(MainWindow.this, "Could not create the neccessary directories for file\n" +
					      "\"" + outFile + "\"!\n",
					      "Error.", JOptionPane.ERROR_MESSAGE);
	} catch(IOException ioe) {
	    ioe.printStackTrace();
	    JOptionPane.showMessageDialog(MainWindow.this, "An exception occurred while extracting to \"" + 
					  outFile + "\": " + ioe.toString() + " (see the console for details)",
					  "Error.", JOptionPane.ERROR_MESSAGE);
	}

    }

    private void extractCompressed(RARFileEntry entry, File outDir, ProgressDialog progressDialog, boolean[] breakSignal) {
// 	System.err.println("Attempting extraction of compressed file.");
// 	NewFileHeader curHead = entry.getHeader(0);
// 	File outFile = new File(outDir, curHead.getFilenameAsString());
// 	try {
// 	    mainPanel.statusArea.append("Unpacking to \"" + outFile + "\"...");
// 	    boolean dirCreated = true;
// 	    if(!outFile.getParentFile().exists())
// 		dirCreated = outFile.getParentFile().mkdirs();
// 	    else if(!outFile.getParentFile().isDirectory())
// 		dirCreated = false;
	    
// 	    if(dirCreated) {
// 		RARFileEntryStream in = new RARFileEntryStream(entry);
// 		FileOutputStream out = new FileOutputStream(outFile);
// 		long bytesToRead = entry.getPackedSize();
// 		long totalBytesRead = 0;
// 		CRC32 crc = new CRC32();
// 		//byte[] buffer = new byte[4096];
// 		//int bytesRead = in.read(buffer);
		
// 		byte[] unpackBuffer = new byte[0x100000];
// 		DecompressionCode dc = new DecompressionCode();
// 		if(curHead.isEncrypted()) {
// 		    String passwd = JOptionPane.showInputDialog(this, "Enter password:"); // INSECURE. we rely on garbage collection to delete this string from memory. should be an array, and set to 0 when done.
// 		    if(passwd != null) {
// 			dc.Encryption = true;
// 			dc.SetCryptKeys(passwd.getBytes("IBM850"));
// 		    }
// 		} 
// 		dc.Unpack(unpackBuffer, in, out, curHead);

// // 	        while(bytesRead != -1 && !breakSignal[0]) {
// // 		    totalBytesRead += bytesRead;
// // 		    progressDialog.setProgress(((double)totalBytesRead)/bytesToRead);
// // 		    crc.update(buffer, 0, bytesRead);
// // 		    out.write(buffer, 0, bytesRead);
// // 		    bytesRead = in.read(buffer);
// // 	        }
// 		out.close();
// 		in.close();
// 		if(false /*breakSignal[0]*/) {
// 		    JOptionPane.showMessageDialog(this, "The file has been partially extracted. Please remove " +
// 						  "it manually.", "Error.", JOptionPane.ERROR_MESSAGE);
// 		}
// 		else {
// 		    if((int)crc.getValue() == (entry.getFinalFileCRC()/* & 0xFFFFFFFFL*/))
// 			mainPanel.statusArea.append("done!\n");
// 		    else {
// 			JOptionPane.showMessageDialog(MainWindow.this, "CRC-32 check failed while extracting to \"" + 
// 						      outFile + "\"!\n" +
// 						      "The file has been extracted but is probably corrupted.\n" + 
// 						      "Calculated CRC-32: 0x" +
// 						      Util.toHexStringBE((int)crc.getValue()) + "\n" +
// 						      "File CRC-32: 0x" + Util.toHexStringBE(curHead.getFileCRC()), 
// 						      "Error.", JOptionPane.ERROR_MESSAGE);
// 			mainPanel.statusArea.append("failed!\n");
// 		    }
// 		}
// 	    }
// 	    else
// 		JOptionPane.showMessageDialog(MainWindow.this, "Could not create the neccessary directories for file\n" +
// 					      "\"" + outFile + "\"!\n",
// 					      "Error.", JOptionPane.ERROR_MESSAGE);
// 	} catch(IOException ioe) {
// 	    ioe.printStackTrace();
// 	    JOptionPane.showMessageDialog(MainWindow.this, "An exception occurred while extracting to \"" + 
// 					  outFile + "\": " + ioe.toString() + " (see the console for details)",
// 					  "Error.", JOptionPane.ERROR_MESSAGE);
// 	}

    }

    public TableModel createNewTableModel(TableModel old, RARFileEntry[] entries) {
	Vector<String> colNames = new Vector<String>(old.getColumnCount());
	for(int i = 0; i < old.getColumnCount(); ++i) {
	    colNames.add(old.getColumnName(i));
	}
	Vector<Vector<String>> rowData = new Vector<Vector<String>>(entries.length);
	for(int i = 0; i < entries.length; ++i) {
	    Vector<String> currentRow = new Vector<String>(colNames.size());
	    NewFileHeader nfh = entries[i].getHeader(0);
	    currentRow.add(nfh.getFilenameAsString());
	    currentRow.add(nfh.getUnpSize() + "");
	    currentRow.add(entries[i].getPackedSize() + "");
	    currentRow.add(Util.toHexStringBE(entries[i].getFinalFileCRC()));
	    
	    while(currentRow.size() < colNames.size())
		currentRow.add("");
	    rowData.add(currentRow);
	}
	
	return new DefaultTableModel(rowData, colNames);
    }

    public void loadFile(File rarFile) {
	try {
	    mainPanel.statusArea.append("Reading contents of file: \"" + rarFile + "\"\n");
	    System.out.println("Reading contents of file: \"" + rarFile + "\"\n");
	    entries = ListFilesInArchive.listFiles(rarFile);
	    int i = 0;
	    for(RARFileEntry entry : entries) {
		NewFileHeader header = entry.getHeader(0);
		if(false) {
		mainPanel.statusArea.append(" Entry " + i++ + ":\n");
		mainPanel.statusArea.append("  Filename : \"" + header.getFilenameAsString() + "\"\n");
// 		mainPanel.statusArea.append("  Unicode filename" + (unicodeFilename == null?" not found.":": \"" + unicodeFilename + "\"") + "\n");
		mainPanel.statusArea.append("  RAR Version: " + (header.getUnpVer()/10) + "." + (header.getUnpVer()%10) + "\n");
		mainPanel.statusArea.append("  method: 0x" + Util.toHexStringBE(header.getMethod()));
		mainPanel.statusArea.append("  flags: 0x" + Util.toHexStringBE(header.getFlags()) + "\n");
		mainPanel.statusArea.append("  attributes: 0x" + Util.toHexStringBE(header.getFileAttr()) + "\n");
		FileAttributes structAttr = header.getFileAttributesStructured();
		mainPanel.statusArea.append("  isDirectory(): " + structAttr.isDirectory() + "\n");
		try {
		    ByteArrayOutputStream baos = new ByteArrayOutputStream();
		    PrintStream ps = new PrintStream(baos, true, "UTF-8");
		    structAttr.print(ps, "  ");
		    mainPanel.statusArea.append(new String(baos.toByteArray(), "UTF-8"));
		} catch(UnsupportedEncodingException uee) {
		    uee.printStackTrace();
		}
		}
		else {
		try {
		    ByteArrayOutputStream baos = new ByteArrayOutputStream();
		    PrintStream ps = new PrintStream(baos, true, "UTF-8");
		    header.print(ps, "  ");
		    mainPanel.statusArea.append(new String(baos.toByteArray(), "UTF-8"));
		} catch(UnsupportedEncodingException uee) {
		    uee.printStackTrace();
		}
		}
	    }
	    mainPanel.statusArea.append("Done.\n\n");
	    mainPanel.fileEntryTable.setModel(createNewTableModel(mainPanel.fileEntryTable.getModel(), entries));
	    setTitle(WINDOW_TITLE + " - [" + rarFile.getName() + "]");
	} catch(Exception e) {
	    e.printStackTrace();
	    
	}
    }
}
